/*
 * Iterators.java
 *
 * Copyright (C) 2007  Neal M Gafter
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA.
 */

package com.gafter;

/**
 * A utility class providing static methods for converting between
 * internal and external iterators.
 * @author gafter
 */
public class Iterators {
    private Iterators() {}
    
    public static <T> InternalIterable<T>
    internalize(final ExternalIterable<T> ext) {
        if (ext == null) throw null;
        return new InternalIterable<T>() {
            public void iterate(Function<T> closure) {
                for (ExternalIterator<T> it = ext.iterator(); it.hasNext(); ) {
                    closure.invoke(it.next());
                }
            }
        };
    }
    
    public static <T> ExternalIterable<T>
    externalize(final InternalIterable<T> it) {
        if (it == null) throw null;
        return new ExternalIterable<T>() {
            public ExternalIterator<T> iterator() {
                class Generator implements Runnable {
                    private boolean needDatum = false;
                    private boolean haveDatum = false;
                    private boolean finished = false;
                    private T datum = null;
                    private final Object lock = new Object();
                    private RuntimeException exitException = null;
                    private Throwable savedThrowable = null;
                    
                    private void waitNeeded() {
                        while (!needDatum && !finished) {
                            try {
                                lock.wait();
                            } catch (InterruptedException ex) {
                                finished = true;
                            }
                        }
                        if (finished) {
                            throw exitException = new RuntimeException();
                        }
                    }
                    public void run() {
                        synchronized(lock) {
                            try {
                                waitNeeded();
                                it.iterate(new Function<T>() {
                                    public void invoke(T t) {
                                        haveDatum = true;
                                        datum = t;
                                        needDatum = false;
                                        lock.notifyAll();
                                        waitNeeded();
                                    }
                                });
                            } catch (Throwable ex) {
                                // an unchecked exception
                                if (ex != exitException) {
                                    savedThrowable = ex;
                                }
                            } finally {
                                finished = true;
                                lock.notifyAll();
                            }
                        }
                    }
                    boolean hasNext() {
                        synchronized (lock) {
                            boolean interrupted = false;
                            needDatum = true;
                            lock.notifyAll();
                            while (!haveDatum && !finished) {
                                try {
                                    lock.wait();
                                } catch (InterruptedException ex) {
                                    interrupted = true;
                                }
                            }
                            if (interrupted) {
                                Thread.currentThread().interrupt();
                            }
                            if (savedThrowable != null) {
                                try {
                                    if (savedThrowable instanceof RuntimeException)
                                        throw (RuntimeException) savedThrowable;
                                    if (savedThrowable instanceof Error)
                                        throw (Error) savedThrowable;
                                    throw new AssertionError(savedThrowable);
                                } finally {
                                    savedThrowable = null;
                                }
                            }
                            return haveDatum;
                        }
                    }
                    T next() {
                        synchronized(lock) {
                            if (hasNext()) {
                                try {
                                    return datum;
                                } finally {
                                    needDatum = false;
                                    haveDatum = false;
                                    datum = null;
                                }
                            } else {
                                throw new java.util.NoSuchElementException();
                            }
                        }
                    }
                    void cleanup() {
                        synchronized(lock) {
                            finished = true;
                            lock.notifyAll();
                        }
                    }
                }
                return new ExternalIterator<T>() {
                    Generator gen = new Generator();
                    Thread daemon = new Thread(gen);
                    {
                        daemon.setDaemon(true);
                        daemon.start();
                    }
                    public boolean hasNext() {
                        return gen.hasNext();
                    }
                    public T next() {
                        return gen.next();
                    }
                    // When no threads reference this iterator object, the
                    // generator thread can be killed.
                    @Override public void finalize() {
                        gen.cleanup();
                        daemon.interrupt();
                    }
                };
            }
        };
    }
}
