/*
 * IteratorsTest.java
 * JUnit based test
 *
 * Created on July 29, 2007, 3:14 PM
 */

package com.gafter;

import java.util.*;
import junit.framework.*;

/**
 *
 * @author gafter
 */
public class IteratorsTest extends TestCase {
    
    public IteratorsTest(String testName) {
        super(testName);
    }
    
    protected void setUp() throws Exception {
    }
    
    protected void tearDown() throws Exception {
    }
    
    /**
     * Test of internalize method, of class com.gafter.Iterators.
     */
    public static void testInternalize0() {
        System.out.println("internalize0");
        
        // ["result0", "result1", null, "result3", "result4"]
        ExternalIterable<String> ext = new ExternalIterable<String>() {
            public ExternalIterator<String> iterator() {
                return new ExternalIterator<String>() {
                    int nextCount = 0;
                    public boolean hasNext() {
                        return nextCount < 5;
                    }
                    public String next() {
                        int k = nextCount++;
                        if (k >= 5) throw new NoSuchElementException();
                        return (k == 2) ? null : "result" + k;
                    }
                };
            }
        };
        
        InternalIterable<String> result = Iterators.internalize(ext);
        result.iterate(new Function<String>() {
            int count = 0;
            public void invoke(String element) {
                int k = count++;
                assertEquals(element, (k == 2) ? null : "result" + k);
            }
        });
    }
    
    /**
     * Test of externalize method, of class com.gafter.Iterators.
     */
    public static void testExternalizeClean() {
        System.out.println("externalizeClean");
        
        InternalIterable<String> it = new InternalIterable<String>() {
            public void iterate(Function<String> closure) {
                for (int i = 0; i < 5; i++) {
                    closure.invoke((i == 2) ? null : "result" + i);
                }
            }
        };
        ExternalIterable<String> result = Iterators.externalize(it);
        int count = 0;
        for (ExternalIterator<String> i = result.iterator(); i.hasNext(); ) {
            int k = count++;
            String element = i.next();
            assertEquals(element, (k == 2) ? null : "result" + k);
        }
        assertEquals(count, 5);
    }
    
    /**
     * Test that the externalize method passes thrown exceptions back
     * to the client.
     */
    public static void testExternalizeThrowing() {
        System.out.println("externalizeThrowing");
        
        InternalIterable<String> it = new InternalIterable<String>() {
            public void iterate(Function<String> closure) {
                for (int i = 0; i < 5; i++) {
                    if (i == 3) throw new RuntimeException("" + i);
                    closure.invoke((i == 2) ? null : "result" + i);
                }
            }
        };
        ExternalIterable<String> result = Iterators.externalize(it);
        int count = 0;
        boolean thrown = false;
        try {
            for (ExternalIterator<String> i = result.iterator(); i.hasNext(); ) {
                int k = count++;
                String element = i.next();
                assertEquals(element, (k == 2) ? null : "result" + k);
            }
        } catch (RuntimeException ex) {
            assertEquals(count, 3);
            thrown = true;
        } finally {
            assertTrue(thrown);
        }
    }
    
    /**
     * Test that the externalize method doesn't invoke more closures than needed.
     */
    public static void testExternalizeLazy0() {
        System.out.println("externalizeLazy0");
        final boolean[] broken = new boolean[1];
        
        InternalIterable<String> it = new InternalIterable<String>() {
            public void iterate(Function<String> closure) {
                broken[0] = true;
            }
        };
        ExternalIterable<String> result = Iterators.externalize(it);
        try { Thread.sleep(10); }
            catch (InterruptedException ex) { Thread.currentThread().interrupt(); }
        assertFalse(broken[0]);
    }
    
    /**
     * Test that the externalize method doesn't invoke more closures than needed.
     */
    public static void testExternalizeLazy1() {
        System.out.println("externalizeLazy0");
        final boolean[] broken = new boolean[1];
        
        InternalIterable<String> it = new InternalIterable<String>() {
            public void iterate(Function<String> closure) {
                for (int i = 0; i < 5; i++) {
                    broken[0] = (i >= 3);
                    closure.invoke((i == 2) ? null : "result" + i);
                }
            }
        };
        ExternalIterable<String> result = Iterators.externalize(it);
        int count = 0;
        for (ExternalIterator<String> i = result.iterator(); i.hasNext(); ) {
            int k = count++;
            String element = i.next();
            assertEquals(element, (k == 2) ? null : "result" + k);
            if (k == 2) break;
        }
        assertFalse(broken[0]);
        assertEquals(count, 3);
    }

    /**
     * Test that the iterator does not start being used unless
     * demanded by the client.
     */
    public static void testExternalizeNoStart() {
        System.out.println("externalizeNoStart");
        final boolean[] started = new boolean[1];
        
        InternalIterable<String> it = new InternalIterable<String>() {
            public void iterate(Function<String> closure) {
                started[0] = true;
            }
        };
        ExternalIterable<String> result = Iterators.externalize(it);
        ExternalIterator<String> iterator = result.iterator();
        try { Thread.sleep(10); }
            catch (InterruptedException ex) { Thread.currentThread().interrupt(); }
        assertFalse(started[0]);
    }

    /**
     * Test that the iterator exits when abandoned. Note that this test
     * may be flaky, since it depends on unspecified timing of garbage
     * collection and finalizers.
     */
    public static void testExternalizeEarlyExit() {
        System.out.println("externalizeEarlyExit");
        final boolean[] completed = new boolean[1];
        
        InternalIterable<String> it = new InternalIterable<String>() {
            public void iterate(Function<String> closure) {
                try {
                    for (int i = 0; i < 5; i++) {
                        closure.invoke((i == 2) ? null : "result" + i);
                    }
                } finally {
                    completed[0] = true;
                }
            }
        };
        ExternalIterable<String> result = Iterators.externalize(it);
        ExternalIterator<String> iterator = result.iterator();
        String x = iterator.next();
        iterator = null;
        System.gc();
        try { Thread.sleep(10); }
            catch (InterruptedException ex) { Thread.currentThread().interrupt(); }
        assertTrue(completed[0]);
    }
}
